'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var path = require('path');
var http = require('http');
var vite = require('vite');
var resolveFrom = require('resolve-from');
var consola = require('consola');
var fs = require('fs-extra');
var fs$1 = require('fs');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

function _interopNamespace(e) {
  if (e && e.__esModule) return e;
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () {
            return e[k];
          }
        });
      }
    });
  }
  n['default'] = e;
  return Object.freeze(n);
}

var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var http__default = /*#__PURE__*/_interopDefaultLegacy(http);
var resolveFrom__default = /*#__PURE__*/_interopDefaultLegacy(resolveFrom);
var consola__default = /*#__PURE__*/_interopDefaultLegacy(consola);
var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var fs__default$1 = /*#__PURE__*/_interopDefaultLegacy(fs$1);

const resolveFile = async (files, dir = ".") => {
  for (const file of files) {
    const absolutePath = path__default['default'].resolve(dir, file);
    if (await fs__default['default'].pathExists(absolutePath)) {
      return absolutePath;
    }
  }
};

const SUPPORTED_CONFIG_FILES = [
  "ream.config.js",
  "ream.config.ts",
  "ream.config.mjs",
  "ream.config.cjs"
];
async function loadConfig(cwd) {
  const file = await resolveFile(SUPPORTED_CONFIG_FILES, cwd);
  if (!file)
    return {};
  const resolved = await vite.loadConfigFromFile({command: "build", mode: "production"}, file);
  return resolved || {};
}

const getInitialState = () => {
  const state = {
    constants: {}
  };
  return state;
};

function normalizePath(path) {
  return path.replace(/\\/g, "/");
}

const removeExt = (p) => p.replace(/\.[a-z0-9]+$/i, "");
class RoutesLoader {
  constructor(dir, fs = fs__default$1['default']) {
    this.dir = dir;
    this.fs = fs;
    this.extRe = /\.(vue|ts|tsx|js|jsx)$/;
  }
  readDir(dir) {
    return this.fs.readdirSync(dir).map((name) => {
      const file = normalizePath(path__default['default'].join(dir, name));
      const stat = this.fs.statSync(file);
      const relativePath = normalizePath(path__default['default'].relative(this.dir, file));
      return {
        path: file,
        relativePath,
        basename: removeExt(name),
        slug: removeExt(relativePath),
        isDir: stat.isDirectory()
      };
    });
  }
  walk(dir, parentPath, endpoints = []) {
    const files = this.readDir(dir);
    let errorFile;
    let notFoundFile;
    let layout;
    let pages = [];
    files.forEach((file) => {
      if (!file.isDir && !this.extRe.test(file.relativePath))
        return;
      if (file.slug === "_error") {
        errorFile = file.path;
        return;
      }
      if (file.slug === "404") {
        notFoundFile = file.path;
        return;
      }
      const isLayout = file.basename === "_layout";
      if (!isLayout && file.basename[0] === "_")
        return;
      let path2 = file.basename === "index" || isLayout ? "" : file.basename;
      path2 = path2.replace(/\[([^\]]+)\]/g, ":$1").replace(/\[\.{3}([^\]]+)\]/, ":$1");
      if (path2) {
        if (parentPath === "/") {
          path2 = parentPath + path2;
        } else {
          path2 = parentPath + "/" + path2;
        }
      } else {
        path2 = parentPath;
      }
      if (file.isDir) {
        this.walk(file.path, path2, endpoints).pages.forEach((page) => pages.push(page));
        return;
      }
      const route = {
        name: file.slug,
        path: path2,
        isEndpoint: /\.[jt]s$/.test(file.relativePath),
        file: file.path
      };
      if (route.isEndpoint) {
        endpoints.push(route);
        return;
      }
      if (file.basename === "_layout") {
        layout = route;
      } else {
        pages.push(route);
      }
    });
    if (layout) {
      layout.children = pages.map((page) => {
        page.path = page.path.replace(layout.path, "").replace(/^\//, "");
        return page;
      });
      pages = [layout];
    }
    return {errorFile, notFoundFile, pages, endpoints};
  }
  load() {
    return this.walk(this.dir, "/");
  }
}
const createRoutesLoader = (dir, fs) => new RoutesLoader(dir, fs);

const isAbsolutPath = (p) => /^\/|[a-zA-Z]:/.test(p);
const writeFileIfChanged = (filepath, content) => {
  if (fs__default['default'].pathExistsSync(filepath)) {
    const prev = fs__default['default'].readFileSync(filepath, "utf8");
    if (prev === content)
      return;
  }
  fs__default['default'].outputFileSync(filepath, content, "utf8");
};
class FileWriter {
  constructor(api) {
    this.api = api;
    this.projectAppHookFiles = [
      this.api.resolveSrcDir("enhance-app.js"),
      this.api.resolveSrcDir("enhance-app.ts")
    ];
    this.projectServerHookFiles = [
      this.api.resolveSrcDir("enhance-server.js"),
      this.api.resolveSrcDir("enhance-server.ts")
    ];
    this.routesLoader = createRoutesLoader(this.routesDir);
  }
  get routesDir() {
    return this.api.resolveSrcDir("routes");
  }
  getRelativePathToTemplatesDir(p) {
    if (!isAbsolutPath(p)) {
      return p;
    }
    return normalizePath(path__default['default'].relative(this.api.resolveDotReam("templates"), p));
  }
  async writeRoutes() {
    const routesInfo = this.routesLoader.load();
    const extendPages = this.api.config.pages;
    if (extendPages) {
      consola__default['default'].info(`Loading extra pages`);
    }
    const pages = extendPages ? await extendPages(routesInfo.pages) : routesInfo.pages;
    const extendEndpoints = this.api.config.endpoints;
    if (extendEndpoints) {
      consola__default['default'].info(`Loading extra endpoints`);
    }
    const endpoints = extendEndpoints ? await extendEndpoints(routesInfo.endpoints) : routesInfo.endpoints;
    const stringifyPages = (pages2) => {
      return `[
        ${pages2.map((page) => {
        return `{
            path: "${page.path}",
            name: "${page.name}",
            meta: {},
            component: function() {
              return import("${this.getRelativePathToTemplatesDir(page.file)}")
                .then(wrapPage)
            },
            ${page.children && page.children.length > 0 ? `children: ${stringifyPages(page.children)}` : ``}
          }`;
      }).join(",")}
      ]`;
    };
    const stringifyEndpoints = (routes) => {
      const serverRoutes = routes.filter((route) => route.isEndpoint);
      return `[
        ${serverRoutes.map((route) => {
        return `{
            path: "${route.path}",
            meta: {load: () => import("${this.getRelativePathToTemplatesDir(route.file)}")},
            component: {}
          }`;
      }).join(",")}${serverRoutes.length === 0 ? "" : ","}
          {
            name: '404',
            path: '/:404(.*)',
            component: {}
          }
      ]`;
    };
    const sharedExportsContent = `
    import { h, defineAsyncComponent } from 'vue'
    import { 
      ErrorComponent as DefaultErrorComponent , 
      NotFoundComponent as DefaultNotFoundComponent } from 'ream/app'
  
    export var ErrorComponent = defineAsyncComponent(function() {
      return ${routesInfo.errorFile ? `import("${this.getRelativePathToTemplatesDir(routesInfo.errorFile)}")` : `Promise.resolve(DefaultErrorComponent)`}
    })
  
    export var NotFoundComponent = defineAsyncComponent(function() {
      return ${routesInfo.notFoundFile ? `import("${this.getRelativePathToTemplatesDir(routesInfo.notFoundFile)}")` : `Promise.resolve(DefaultNotFoundComponent)`}
    })
  
    var wrapPage = function(page) {
      var Component = page.default
      return {
        ...Component,
        $$load: page.load,
        $$preload: page.preload,
        $$getStaticPaths: page.getStaticPaths,
        $$transition: page.transition,
      }
    }
  
    var clientRoutes = ${stringifyPages(pages)}

    clientRoutes.push({
      name: '404',
      path: '/:404(.*)',
      meta:{},
      component: import.meta.env.DEV ? {
        render() {
          return h('h1','error: this component should not be rendered')
        }
      } : {}
    })
  
    export {
      clientRoutes,
    }
    `;
    writeFileIfChanged(this.api.resolveDotReam("templates/shared-exports.js"), sharedExportsContent);
    const serverExportsContent = `
   export const serverRoutes = ${stringifyEndpoints(endpoints)}
    `;
    writeFileIfChanged(this.api.resolveDotReam("templates/server-exports.js"), serverExportsContent);
  }
  async writeGlobalImports() {
    writeFileIfChanged(this.api.resolveDotReam("templates/global-imports.js"), `
      ${this.api.config.imports.map((file) => `import ${JSON.stringify(file)}`).join("\n")}
      `);
  }
  async writeIndexHtml() {
    const filepath = this.api.resolveSrcDir("index.html");
    if (await fs__default['default'].pathExists(filepath))
      return;
    const content = `<!DOCTYPE html>
<html ream-html-attrs>
<head>
  <meta charset="UTF-8">
  <meta http-equiv="X-UA-Compatible" content="IE=edge">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <!--ream-head-->
</head>
<body ream-body-attrs>
  <!--ream-main-->
</body>
</html>`;
    await fs__default['default'].outputFile(filepath, content, "utf8");
  }
  async writeHookFile(type) {
    let files = [];
    for (const plugin of this.api.config.plugins) {
      if (type === "app") {
        if (plugin.enhanceAppFiles) {
          files = [...files, ...plugin.enhanceAppFiles.call(this.api)];
        }
      } else if (type === "server") {
        if (plugin.enhanceServerFiles) {
          files = [...files, ...plugin.enhanceServerFiles.call(this.api)];
        }
      }
    }
    const projectFile = await resolveFile(type === "app" ? this.projectAppHookFiles : this.projectServerHookFiles);
    if (projectFile) {
      files.push(projectFile);
    }
    let content = `
    ${files.map((file, index) => {
      return `import * as hook_${type}_${index} from "${this.getRelativePathToTemplatesDir(file)}"`;
    }).join("\n")}
  
  var files = [
    ${files.map((_, i) => `hook_${type}_${i}`).join(",")}
  ]
  
  function getExportByName(name) {
    var fns = []
    for (var i = 0; i < files.length; i++) {
      var mod = files[i]
      if (mod[name]) {
        fns.push(mod[name])
      }
    }
    return fns
  }
  
  export async function callAsync(name, context) {
    for (const fn of getExportByName(name)) {
      await fn(context)
    }
  }
  `;
    if (type === "server") {
      content += `
    export const hasExport = (name) => getExportByName(name).length > 0
    `;
    }
    writeFileIfChanged(this.api.resolveDotReam(`templates/enhance-${type}.js`), content);
  }
}
const preparePlugin = () => {
  let writer;
  return {
    name: `ream:prepare`,
    async prepare() {
      writer = new FileWriter(this);
      if (!await fs__default['default'].pathExists(writer.routesDir)) {
        throw new Error(`${writer.routesDir} doesn't exist`);
      }
      await Promise.all([
        writer.writeRoutes(),
        writer.writeHookFile("app"),
        writer.writeHookFile("server"),
        writer.writeGlobalImports(),
        writer.writeIndexHtml()
      ]);
    },
    async onFileChange(type, file) {
      if (file.startsWith(writer.routesDir) && writer.routesLoader.extRe.test(file)) {
        if (type === "add" || type === "unlink") {
          await writer.writeRoutes();
        }
      }
      if (writer.projectAppHookFiles.includes(file)) {
        await writer.writeHookFile("app");
      }
      if (writer.projectServerHookFiles.includes(file)) {
        await writer.writeHookFile("server");
      }
    }
  };
};

const getPort = async (host = "localhost", port) => {
  if (!port) {
    throw new Error(`You need to provide "port" option when initializing Ream`);
  }
  process.env.PORT = String(port);
  process.env.HOST = host;
  return {host, port};
};

const defineReamConfig = (config) => config;
const OWN_DIR = path__default['default'].join(__dirname, "../");
const resolveOwnDir = (...args) => path__default['default'].resolve(OWN_DIR, ...args);
class Ream {
  constructor(options, inlineConfig = {}) {
    this.inlineConfig = inlineConfig;
    this.rootDir = path__default['default'].resolve(options.rootDir || ".");
    this.host = options.host;
    this.port = options.port;
    if (options.srcDir) {
      this.srcDir = path__default['default'].join(this.rootDir, options.srcDir);
    } else {
      const hasRoutesInSrc = fs__default['default'].existsSync(path__default['default'].join(this.rootDir, "src/routes"));
      this.srcDir = hasRoutesInSrc ? path__default['default'].join(this.rootDir, "src") : this.rootDir;
    }
    this.isDev = Boolean(options.dev);
    this.mode = options.mode || (this.isDev ? "development" : "production");
    if (!process.env.NODE_ENV) {
      process.env.NODE_ENV = this.isDev ? "development" : "production";
    }
  }
  resolveRootDir(...args) {
    return path__default['default'].resolve(this.rootDir, ...args);
  }
  resolveSrcDir(...args) {
    return path__default['default'].resolve(this.srcDir, ...args);
  }
  resolveDotReam(...args) {
    return this.resolveRootDir(".ream", ...args);
  }
  resolveInPackage(pkg, target) {
    const pkgDir = path__default['default'].dirname(resolveFrom__default['default'](process.cwd(), `${pkg}/package.json`));
    return resolveFrom__default['default'](pkgDir, target);
  }
  async loadConfig() {
    const {config = {}, path: configPath} = await loadConfig(this.rootDir);
    if (configPath) {
      consola__default['default'].info(`Using config file: ${path__default['default'].relative(process.cwd(), configPath)}`);
    }
    this.configPath = configPath;
    this.config = {
      ...this.inlineConfig,
      ...config,
      env: {
        ...config.env,
        ...this.inlineConfig.env
      },
      plugins: [
        preparePlugin(),
        ...this.inlineConfig.plugins || [],
        ...config.plugins || []
      ],
      imports: config.imports || []
    };
  }
  get env() {
    return {
      ...this.userEnv,
      ...this.config.env
    };
  }
  get constants() {
    const {env} = this;
    const srcDir = path__default['default'].relative(process.cwd(), this.resolveSrcDir()) || ".";
    const rootDir = path__default['default'].relative(process.cwd(), this.resolveRootDir()) || ".";
    return {
      ...this.state.constants,
      ...Object.keys(env).reduce((res, key) => {
        const value = JSON.stringify(env[key]);
        return {
          ...res,
          [`import.meta.env.${key}`]: value,
          [`process.env.${key}`]: value
        };
      }, {}),
      REAM_SOURCE_DIR: JSON.stringify(srcDir),
      REAM_ROOT_DIR: JSON.stringify(rootDir)
    };
  }
  ensureEnv(name, defaultValue) {
    const value = this.env[name];
    if (!value && !defaultValue) {
      throw new Error(`Environment variable "${name}" is required`);
    }
    return value || defaultValue;
  }
  async prepare(type) {
    this.state = getInitialState();
    if (type !== "build") {
      const {port, host} = await getPort(this.host, this.port);
      this.host = host;
      this.port = port;
    }
    if (type !== "start") {
      await this.loadConfig();
      await Promise.all(["templates", "manifest", "server", "client", "export"].map((name) => {
        return fs__default['default'].remove(this.resolveDotReam(name));
      }));
    }
    this.userEnv = vite.loadEnv(this.mode, this.rootDir, "REAM_");
    if (type !== "start") {
      for (const plugin of this.config.plugins) {
        if (plugin.prepare) {
          await plugin.prepare.call(this);
        }
      }
    }
    if (type === "dev") {
      this._devServer = http__default['default'].createServer();
      const {createServer: createViteServer} = await Promise.resolve().then(function () { return /*#__PURE__*/_interopNamespace(require('vite')); });
      const {getViteConfig} = await Promise.resolve().then(function () { return require('./get-vite-config-60ced8ef.js'); });
      const viteConfig = getViteConfig(this);
      const viteDevServer = await createViteServer(viteConfig);
      if (this.viteDevServer) {
        for (const key in viteDevServer) {
          if (key !== "app") {
            this.viteDevServer[key] = viteDevServer[key];
          }
        }
      } else {
        this.viteDevServer = viteDevServer;
      }
      viteDevServer.watcher.on("all", async (_, file) => {
        const files = SUPPORTED_CONFIG_FILES.map((name) => this.resolveRootDir(name));
        if (file === this.configPath || files.includes(file)) {
          consola__default['default'].info(`Restarting Ream due to changes in ${file}`);
          await viteDevServer.close();
          await this.prepare(type);
        }
      });
      for (const plugin of this.config.plugins) {
        if (plugin.onFileChange) {
          viteDevServer.watcher.on("all", plugin.onFileChange);
        }
      }
    }
  }
  localResolve(name) {
    return resolveFrom__default['default'].silent(this.rootDir, name);
  }
  localRequire(name) {
    const path2 = this.localResolve(name);
    return path2 && require(path2);
  }
  async getRequestHandler() {
    await this.prepare(this.isDev ? "dev" : "start");
    if (this.isDev) {
      const {getRequestHandler} = await Promise.resolve().then(function () { return require('./dev-server-22d8808f.js'); });
      return getRequestHandler(this);
    }
    const {createHandler} = await Promise.resolve().then(function () { return require('./server.js'); });
    const serverAssets = await getServerAssets(this.resolveDotReam());
    return createHandler({
      cwd: this.rootDir,
      ...serverAssets
    });
  }
  async startServer() {
    const handler = await this.getRequestHandler();
    const server = this._devServer || http__default['default'].createServer();
    server.on("request", handler);
    server.listen(this.port, this.host);
    console.log(`> Open http://${this.host}:${this.port}`);
  }
  async build({
    fullyExport,
    standalone
  }) {
    await this.prepare("build");
    const {build, buildStandalone} = await Promise.resolve().then(function () { return require('./build-95cdf119.js'); });
    await build(this);
    const {exportSite} = await Promise.resolve().then(function () { return require('./export-f0e40eae.js'); });
    await exportSite(this.resolveDotReam(), fullyExport);
    if (standalone) {
      await buildStandalone(this);
    }
  }
}
const readJSON = (filepath) => JSON.parse(fs__default['default'].readFileSync(filepath, "utf8"));
const getServerAssets = async (dotReamDir) => {
  const serverEntry = await Promise.resolve().then(function () { return /*#__PURE__*/_interopNamespace(require(path__default['default'].join(dotReamDir, "server/server-entry.js"))); });
  const ssrManifest = readJSON(path__default['default'].join(dotReamDir, "manifest/ssr-manifest.json"));
  const clientManifest = readJSON(path__default['default'].join(dotReamDir, "manifest/client-manifest.json"));
  let exportManifest;
  try {
    exportManifest = readJSON(path__default['default'].join(dotReamDir, "manifest/export-manifest.json"));
  } catch (_) {
  }
  const htmlTemplate = fs__default['default'].readFileSync(path__default['default'].join(dotReamDir, "client/index.html"), "utf8");
  return {
    serverEntry,
    ssrManifest,
    clientManifest,
    exportManifest,
    htmlTemplate
  };
};

exports.OWN_DIR = OWN_DIR;
exports.Ream = Ream;
exports.defineReamConfig = defineReamConfig;
exports.getServerAssets = getServerAssets;
exports.resolveOwnDir = resolveOwnDir;
