'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var path = require('path');
var fs = require('fs');
var serveStatic = require('sirv');
var querystring = require('querystring');
var router = require('@egoist/router');
var serverRenderer = require('@vue/server-renderer');
var serializeJavascript = require('serialize-javascript');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var serveStatic__default = /*#__PURE__*/_interopDefaultLegacy(serveStatic);
var serializeJavascript__default = /*#__PURE__*/_interopDefaultLegacy(serializeJavascript);

const TYPE = "content-type";
const OSTREAM = "application/octet-stream";
function send(res, data = "") {
  let type = res.getHeader("content-type");
  if (data && typeof data.pipe === "function") {
    res.setHeader(TYPE, type || OSTREAM);
    return data.pipe(res);
  }
  if (data instanceof Buffer) {
    type = type || OSTREAM;
  } else if (typeof data === "object") {
    data = JSON.stringify(data);
    type = type || "application/json;charset=utf-8";
  } else {
    type = type || "text/html";
  }
  res.setHeader("content-length", Buffer.byteLength(data));
  res.setHeader("content-type", type);
  res.end(data);
}
function status(res, code) {
  res.statusCode = code;
}

const parseUrl = (url) => {
  const queryIndex = url.indexOf("?");
  if (queryIndex === -1) {
    return {path: url, search: ""};
  }
  const path = url.substring(0, queryIndex);
  const search = url.substring(queryIndex);
  return {path, search};
};
class ConnectError extends Error {
}
const onError = (error, _, res) => {
  const status = typeof error === "string" ? 500 : error.status || 500;
  res.statusCode = status;
  res.end(`${status}`);
};
const notFoundError = new ConnectError("404 not found");
notFoundError.status = 404;
class Connect {
  constructor(options = {}) {
    this.get = this.add.bind(this, "GET");
    this.head = this.add.bind(this, "HEAD");
    this.patch = this.add.bind(this, "PATCH");
    this.options = this.add.bind(this, "OPTIONS");
    this.connect = this.add.bind(this, "CONNECT");
    this.delete = this.add.bind(this, "DELETE");
    this.trace = this.add.bind(this, "TRACE");
    this.post = this.add.bind(this, "POST");
    this.put = this.add.bind(this, "PUT");
    this.handler = (_req, _res) => {
      const req = _req;
      const res = _res;
      const info = parseUrl(req.url);
      req.originalUrl = req.originalUrl || req.url;
      req.path = info.path;
      req.query = querystring.parse(info.search.substring(1));
      const matches = this.router.find(req.method, req.url);
      matches.push({handler: this.onNoMatch, params: {}});
      let i = 0;
      const next = async (error) => {
        if (error)
          return this.onError(error, req, res);
        const m = matches[i++];
        if (m) {
          req.params = m.params;
          try {
            await m.handler(req, res, next);
          } catch (error2) {
            next(error2);
          }
        }
      };
      next();
    };
    this.router = new router.Router();
    this.onError = options.onError || onError;
    this.onNoMatch = options.onNoMatch || this.onError.bind(null, notFoundError);
  }
  add(method, route, ...handlers) {
    this.router.add(method, route, ...handlers);
    return this;
  }
  use(route, ...handlers) {
    if (typeof route === "string") {
      this.router.use(route, ...handlers);
    } else if (typeof route === "function") {
      this.router.use("/*", route, ...handlers);
    }
    return this;
  }
}
const connect = (options) => new Connect(options);

const createServer = (options) => {
  const app = connect(options);
  app.use((req, res, next) => {
    req.isLoadRequest = req.path.endsWith(".load.json");
    Object.defineProperties(res, {
      send: {
        enumerable: true,
        value: function(data) {
          send(res, data);
          return res;
        }
      },
      status: {
        enumerable: true,
        value: function(statusCode) {
          status(res, statusCode);
          return res;
        }
      }
    });
    next();
  });
  return app;
};

function renderPreloadLinks(modules, manifest) {
  let links = "";
  if (!manifest) {
    return links;
  }
  const seen = new Set();
  modules.forEach((id) => {
    const files = manifest[id];
    if (files) {
      files.forEach((file) => {
        if (!seen.has(file)) {
          seen.add(file);
          links += renderPreloadLink(file);
        }
      });
    }
  });
  return links;
}
function renderPreloadLink(file) {
  if (file.endsWith(".js")) {
    return `<link rel="modulepreload" crossorigin href="${file}">`;
  } else if (file.endsWith(".css")) {
    return `<link rel="stylesheet" href="${file}">`;
  } else {
    return "";
  }
}
async function render({
  url,
  loadOptions,
  ssrManifest,
  serverEntry,
  isLoadRequest,
  clientManifest,
  router,
  notFound,
  htmlTemplate
}) {
  let html;
  let loadResult;
  const route = router.currentRoute.value;
  const components = route.matched.map((route2) => route2.components.default);
  loadResult = await loadPageData(void 0, components, loadOptions);
  if (notFound || route.name === "404") {
    loadResult.notFound = true;
  }
  if (!isLoadRequest) {
    html = await renderToHTML({
      url,
      path: route.path,
      loadResult,
      serverEntry,
      router,
      ssrManifest,
      clientManifest,
      htmlTemplate
    });
  }
  return {html, loadResult};
}
async function renderToHTML(options) {
  const context = {
    url: options.url,
    initialState: {},
    router: options.router
  };
  context.initialState = {
    load: {
      [options.path]: options.loadResult
    }
  };
  const app = await options.serverEntry.render(context);
  const appHTML = await serverRenderer.renderToString(app, context);
  const preloadLinks = context.modules ? renderPreloadLinks(context.modules, options.ssrManifest) : "";
  const headHTML = options.serverEntry.renderHead(app);
  const html = options.htmlTemplate.replace(" ream-html-attrs", headHTML.htmlAttrs).replace(" ream-body-attrs", headHTML.bodyAttrs).replace("<!--ream-head-->", `${headHTML.headTags}${preloadLinks}`).replace("<!--ream-main-->", `<div id="_ream">${appHTML}</div>
    <script>INITIAL_STATE=${serializeJavascript__default['default'](context.initialState, {
    isJSON: true
  })}
    </script>
    `);
  return html;
}
async function loadPageData(globalLoad, components, options) {
  const props = {};
  let hasPreload;
  let hasLoad;
  let notFound;
  let error;
  let redirect;
  let revalidate;
  const fns = globalLoad ? [globalLoad] : [];
  for (const component of components) {
    if (component.$$load) {
      hasLoad = true;
    }
    if (component.$$preload) {
      hasPreload = true;
    }
    const load = component.$$load || component.$$preload;
    if (load) {
      fns.push(load);
    }
  }
  if (fns.length > 0) {
    for (const fn of fns) {
      const result = await fn(options);
      if (result) {
        if (typeof result.revalidate === "number") {
          revalidate = result.revalidate * 1e3;
        }
        if (result.props) {
          Object.assign(props, result.props);
          continue;
        }
        if (result.notFound) {
          notFound = true;
        } else if (result.redirect) {
          redirect = typeof result.redirect === "string" ? {url: result.redirect} : result.redirect;
        } else if (result.error) {
          error = result.error;
        }
        break;
      } else if (result == null) {
        notFound = true;
        break;
      }
    }
  }
  return {
    props,
    hasPreload,
    hasLoad,
    notFound,
    error,
    redirect,
    revalidate,
    expiry: revalidate && revalidate + Date.now()
  };
}

class LRU {
  constructor(options = {}) {
    this.first = null;
    this.items = Object.create(null);
    this.last = null;
    this.max = options.max || 0;
    this.size = 0;
    this.ttl = options.ttl || 0;
  }
  has(key) {
    return this.items[key] !== void 0;
  }
  clear() {
    this.first = null;
    this.items = Object.create(null);
    this.last = null;
    this.size = 0;
  }
  delete(key) {
    if (this.has(key)) {
      const item = this.items[key];
      delete this.items[key];
      this.size--;
      if (item.prev !== null) {
        item.prev.next = item.next;
      }
      if (item.next !== null) {
        item.next.prev = item.prev;
      }
      if (this.first === item) {
        this.first = item.next;
      }
      if (this.last === item) {
        this.last = item.prev;
      }
    }
  }
  evict() {
    if (this.first) {
      const key = this.first.key;
      this.first = this.first.next;
      if (this.first) {
        this.first.prev = null;
      }
      delete this.items[key];
      this.size--;
    }
  }
  get(key) {
    const item = this.items[key];
    if (item) {
      if (item.ttl > 0 && item.expiry <= Date.now()) {
        this.delete(key);
      } else {
        this.set(key, item.value, item.ttl, true);
        return item.value;
      }
    }
  }
  keys() {
    return Object.keys(this.items);
  }
  set(key, value, ttl = this.ttl, bypassChecking = false) {
    let item = this.items[key];
    if (bypassChecking || item !== void 0) {
      item.value = value;
      if (bypassChecking === false) {
        item.expiry = ttl > 0 ? Date.now() + ttl : ttl;
      }
      if (this.last && this.last !== item) {
        const last = this.last, next = item.next, prev = item.prev;
        if (this.first === item) {
          this.first = item.next;
        }
        item.next = null;
        item.prev = this.last;
        last.next = item;
        if (prev !== null) {
          prev.next = next;
        }
        if (next !== null) {
          next.prev = prev;
        }
      }
    } else {
      if (this.max > 0 && this.size === this.max) {
        this.evict();
      }
      item = {
        ttl,
        expiry: ttl > 0 ? Date.now() + ttl : ttl,
        key,
        prev: this.last,
        next: null,
        value
      };
      this.items[key] = item;
      if (++this.size === 1) {
        this.first = item;
      } else if (this.last) {
        this.last.next = item;
      }
    }
    this.last = item;
  }
}

const outputFile = async (filepath, content, encoding) => {
  await fs__default['default'].promises.mkdir(path__default['default'].dirname(filepath), {recursive: true});
  return fs__default['default'].promises.writeFile(filepath, content, encoding);
};
const readFile = fs__default['default'].promises.readFile;
const pathExists = (filepath) => new Promise((resolve) => {
  fs__default['default'].access(filepath, (err) => resolve(!err));
});

const getExportOutputPath = (pathname, type, exportDir) => {
  let filename;
  pathname = pathname.replace(/\/$/, "") || "index";
  if (type === "json") {
    filename = `${pathname}.load.json`;
  } else {
    if (pathname.endsWith(".html")) {
      filename = pathname;
    } else {
      filename = pathname === "index" ? "index.html" : `${pathname}/index.html`;
    }
  }
  return path__default['default'].join(exportDir, filename);
};
class ExportCache {
  constructor(options) {
    this.cache = new LRU({max: 1e3});
    this.options = options;
  }
  getOutputPath(pathname, type) {
    return getExportOutputPath(pathname, type, this.options.exportDir);
  }
  get writeOnly() {
    return this.options.writeOnly;
  }
  findStaticPage(rawPath) {
    const {exportManifest} = this.options;
    if (!exportManifest)
      return;
    return exportManifest.staticPages.find((record) => record.path === rawPath);
  }
  async get(pathname) {
    if (this.options.writeOnly)
      return;
    let data = this.cache.get(pathname);
    if (!data && this.options.flushToDisk) {
      const htmlPath = this.getOutputPath(pathname, "html");
      const jsonPath = this.getOutputPath(pathname, "json");
      try {
        const [html, loadResult] = await Promise.all([
          pathExists(htmlPath).then((exists) => exists ? readFile(htmlPath, "utf8") : ""),
          pathExists(jsonPath).then((exists) => exists && readFile(jsonPath, "utf8").then((res) => JSON.parse(res)))
        ]);
        if (html || loadResult) {
          data = {html, loadResult};
        }
      } catch (_) {
        console.log(_);
      }
    }
    if (data && data.loadResult.expiry && data.loadResult.expiry <= Date.now()) {
      data.isStale = true;
    }
    return data;
  }
  async set(pathname, {html, loadResult}, flushToDisk = this.options.flushToDisk) {
    this.cache.set(pathname, {html, loadResult});
    if (flushToDisk) {
      const htmlPath = this.getOutputPath(pathname, "html");
      const jsonPath = this.getOutputPath(pathname, "json");
      const json = serializeJavascript__default['default'](loadResult, {isJSON: true});
      await Promise.all([
        html ? outputFile(htmlPath, html, "utf8") : null,
        outputFile(jsonPath, json, "utf8")
      ]);
    }
  }
}

const createClientRouter = async (serverEntry, url) => {
  const router = serverEntry.createClientRouter();
  await serverEntry.enhanceApp.callAsync("onCreatedRouter", {router});
  await router.push(url);
  await router.isReady();
  return router;
};
const getLoadOptions = (req) => {
  return {
    get path() {
      return req.path;
    },
    get params() {
      return req.params;
    },
    get host() {
      return req.headers.host;
    },
    get query() {
      return req.query;
    },
    get headers() {
      return req.headers;
    }
  };
};
async function createHandler(options) {
  const dotReamDir = path__default['default'].resolve(options.cwd || ".", ".ream");
  const serverEntry = options.serverEntry;
  const ssrManifest = options.ssrManifest || {};
  const clientManifest = options.clientManifest || {};
  const htmlTemplate = options.htmlTemplate;
  const exportCache = options.exportManifest && new ExportCache({
    exportDir: path__default['default'].join(dotReamDir, "export"),
    flushToDisk: true,
    exportManifest: options.exportManifest
  });
  const onError = async (err, req, res) => {
    if (typeof err === "string") {
      err = new Error(err);
    }
    if (options.fixStacktrace) {
      options.fixStacktrace(err);
    }
    console.error("server error", err.stack);
    res.statusCode = !res.statusCode || res.statusCode < 400 ? 500 : res.statusCode;
    const error = {
      status: res.statusCode,
      message: options.dev ? err.stack : void 0
    };
    if (req.isLoadRequest) {
      return res.send({error});
    }
    try {
      const router = await createClientRouter(serverEntry, req.url);
      const ErrorComponent = await serverEntry.ErrorComponent.__asyncLoader();
      const loadResult = await loadPageData(void 0, [ErrorComponent], getLoadOptions(req));
      loadResult.error = error;
      const html = await renderToHTML({
        path: req.path,
        url: req.url,
        router,
        loadResult,
        serverEntry,
        ssrManifest,
        clientManifest,
        htmlTemplate
      });
      res.setHeader("content-type", "text-html");
      res.end(html);
    } catch (error2) {
      res.statusCode = 500;
      res.send(options.dev ? `<h1>Error occurs while rendering error page:</h1><pre>${error2.stack}</pre>` : "server error");
    }
  };
  const server = createServer({
    onError
  });
  if (options.before) {
    options.before(server);
  }
  const renderNotFound = async (req, res, router) => {
    res.statusCode = 404;
    if (req.isLoadRequest) {
      return res.send({status: 404});
    }
    let html;
    if (exportCache) {
      const cache = await exportCache.get("/404.html");
      if (cache) {
        html = cache.html || "";
      }
    }
    if (!html) {
      const result = await render({
        url: req.url,
        loadOptions: getLoadOptions(req),
        isLoadRequest: false,
        ssrManifest,
        serverEntry,
        clientManifest,
        router,
        notFound: true,
        htmlTemplate
      });
      html = result.html;
    }
    res.send(html);
  };
  const clientDir = path__default['default'].join(dotReamDir, "client");
  if (!options.dev && fs__default['default'].existsSync(clientDir)) {
    const serveStaticFiles = serveStatic__default['default'](clientDir);
    server.use((req, res, next) => {
      if (req.path === "/" || req.path === "/index.html")
        return next();
      return serveStaticFiles(req, res, next);
    });
  }
  await serverEntry.enhanceServer.callAsync("extendServer", {
    server
  });
  server.use(async function handleServerRoutes(req, res, next) {
    const serverRouter = serverEntry.createServerRouter();
    await serverRouter.push(req.url);
    await serverRouter.isReady();
    const {matched} = serverRouter.currentRoute.value;
    const route = matched[0];
    if (route.name === "404") {
      return next();
    }
    const mod = await route.meta.load();
    return mod.default(req, res, next);
  });
  server.use(async function handleAppRoute(req, res, next) {
    if (req.method !== "GET") {
      return next();
    }
    if (req.isLoadRequest) {
      req.url = req.url.replace(/(\/index)?\.load\.json/, "");
      if (req.url[0] !== "/") {
        req.url = `/${req.url}`;
      }
      const queryIndex = req.url.indexOf("?");
      if (queryIndex === -1) {
        req.path = req.url;
      } else {
        req.path = req.url.substring(0, queryIndex);
      }
    }
    const router = await createClientRouter(serverEntry, req.url);
    const route = router.currentRoute.value;
    req.params = route.params;
    if (route.name === "404") {
      await renderNotFound(req, res, router);
      return;
    }
    const rawPath = route.matched.map((m) => m.path).join("/");
    const staticPage = exportCache == null ? void 0 : exportCache.findStaticPage(rawPath);
    const pageCache = staticPage && await (exportCache == null ? void 0 : exportCache.get(req.path));
    if (pageCache) {
      if (req.isLoadRequest) {
        res.send(serializeJavascript__default['default'](pageCache.loadResult, {isJSON: true}));
      } else if (pageCache.html) {
        res.send(pageCache.html);
      }
      if (res.writableEnded && !pageCache.isStale) {
        return;
      }
    }
    let fallback = true;
    if (!pageCache && staticPage && !staticPage.fallback) {
      fallback = false;
    }
    if (!fallback) {
      await renderNotFound(req, res, router);
      return;
    }
    const {html, loadResult} = await render({
      url: req.url,
      loadOptions: getLoadOptions(req),
      isLoadRequest: req.isLoadRequest,
      ssrManifest,
      serverEntry,
      clientManifest,
      htmlTemplate,
      router
    });
    if (loadResult.error) {
      res.statusCode = loadResult.error.status;
    }
    if (!pageCache || !pageCache.isStale) {
      if (req.isLoadRequest) {
        res.send(serializeJavascript__default['default'](loadResult, {isJSON: true}));
      } else {
        if (loadResult.redirect) {
          res.writeHead(loadResult.redirect.permanent ? 301 : 302, {
            Location: loadResult.redirect.url
          });
          res.end();
        } else {
          res.send(html || "");
        }
      }
    }
    if (exportCache && !loadResult.hasLoad && !options.dev) {
      await exportCache.set(req.path, {
        html,
        loadResult
      });
    }
  });
  return server.handler;
}

exports.Connect = Connect;
exports.ExportCache = ExportCache;
exports.createClientRouter = createClientRouter;
exports.createHandler = createHandler;
exports.getExportOutputPath = getExportOutputPath;
exports.getLoadOptions = getLoadOptions;
exports.loadPageData = loadPageData;
exports.render = render;
exports.renderToHTML = renderToHTML;
